# AUTOGENERATED! DO NOT EDIT! File to edit: ../nbs/18_applications.ipynb.

# %% auto 0
__all__ = ['run_wind_turbine']

# %% ../nbs/18_applications.ipynb 3
from .environment_processing import EnvironmentProcessingFactory


# %% ../nbs/18_applications.ipynb 4
def run_wind_turbine(target, drive, filepath):
    verbosed = {'debug': 0, 'hpct_verbose':0}

    # drive = get_gdrive()
    # root_path=get_root_path()
    root_path=""
    results_props = {'comparisons' : True, 'comparisons_print_plots': False}
    env_name = 'WindTurbine'
    args = {'file': filepath, 'env_name':env_name, 'verbosed':verbosed, 'drive':drive, 'root_path':root_path, 'max' : True} | results_props
    # args = {'file': filepath, 'env_name':env_name, 'verbosed':verbosed, 'max' : True} | results_props

    environment_properties={'series': target, 'zero_threshold': 1, 'reward_type': 'power', 'keep_history': True, 'range': 'test'}

    env_proc = EnvironmentProcessingFactory.createEnvironmentProcessing(f'{env_name}EnvironmentProcessing')
    env_proc.set_properties(args=args)

    env_proc.get_experiment()
    rtn = env_proc.results(filepath=filepath, environment_properties=environment_properties)


